#!/usr/bin/env python

"""
An example of how to stack evenly-spaced N-dimensional spline surfaces
to form an N+1-dimensional spline surface that maximally fills the
convex hull given by the input surfaces.
"""

import numpy
try:
	# Fast sparse-matrix implementation
	from photospline import spglam as glam
except ImportError:
	# Slow (yet readable) pure-python implementation
	from photospline.glam import glam
from photospline.glam.glam import grideval
from photospline.glam.bspline import bspline
from photospline import splinefitstable
import os
from utils import colorize, pad_knots, stack_tables

knots = pad_knots(numpy.linspace(0, 2*numpy.pi, 11))
x = numpy.linspace(0, 2*numpy.pi, 101)
weights = numpy.ones(x.shape)
smooth = 1e-3
order = 2

import pylab
xfine = numpy.linspace(x[0], x[-1], 1001)
scatterstyle = dict(ls='None', capsize=0., fmt='k')
y = numpy.linspace(0, 1, 7)
z = numpy.zeros((x.size, y.size))
splines = []

# generate some random data
for i, (color, yc) in enumerate(colorize(y)):
	zmean = (yc*1000 + 500)*(yc*numpy.cos(x)**2 + (1-yc)*numpy.sin(x)**2 + 2*yc)
	z[:,i] = numpy.random.poisson(zmean)

# fit a spline to each surface individually
for i, yc in enumerate(y):
	spline = glam.fit(z[:,i], weights, [x], [knots], order, smooth)
	splines.append(spline)

label = 'single surfaces'
for i, (color, yc) in enumerate(colorize(y)):
	pylab.errorbar(x, z[:,i], xerr=numpy.diff(x)[0]/2., yerr=numpy.sqrt(z[:,i]), color=color, **scatterstyle)
	pylab.plot(xfine, grideval(splines[i], [xfine]), color=color, label=label)
	label = '_nolabel_'
	# print yc, color

# Now, construct a spline surface that interpolates between the
# individual surfaces. We clone the first and last surfaces so
# that the interpolating surface has full support at the edges.
splines = [splines[0]] + splines + [splines[-1]]
yint = [y[0]-(y[1]-y[0])] + list(y) + [y[-1]+(y[-1]-y[-2])]
interpolant = stack_tables(zip(splines, yint))

# Plot the pseudo-interpolating spline
yfine = numpy.linspace(y[0], y[-1], 2*(y.size-1)+1)
zfine = grideval(interpolant, [xfine, yfine])
label = 'stacked'
for color, i in colorize(range(yfine.size)):
	pylab.plot(xfine, zfine[:,i], color=color, ls=':', label=label)
	label = '_nolabel_'

# For comparison, plot a 2-dimensional surface fit by least-squares
yknots = pad_knots(numpy.linspace(0, 1, 3))
interpolant = glam.fit(z, numpy.ones(z.shape), [x, y], [knots, yknots], [order, order], smooth)
zfine = grideval(interpolant, [xfine, yfine])
label = 'global fit'
for color, i in colorize(range(yfine.size)):
	pylab.plot(xfine, zfine[:,i], color=color, ls='--', label=label)
	label = '_nolabel_'

pylab.legend()
pylab.show()
