#!/usr/bin/env python

"""
An example of how to fit 1-d data to a spline surface and plot the result.
Modify the knot positions, spline order, and smoothing parameter to see the
effect these have on the solution.
"""

import numpy
try:
	# Fast sparse-matrix implementation
	from photospline import spglam as glam
except ImportError:
	# Slow (yet readable) pure-python implementation
	from photospline.glam import glam
from photospline.glam.glam import grideval
from photospline.glam.bspline import bspline
from photospline import splinefitstable
import os

numpts = 500
knots = [numpy.linspace(-8,35,30)]
order = 2
smooth = 3.14159e3

x1 = numpy.sort(numpy.random.uniform(-4,25,size=numpts))
z = numpy.random.poisson(numpy.cos(x1)**2 + (x1-3.0)**2 + 10)
weight = 1.+z
result = glam.fit(z,weight,[x1],knots,order,smooth)

# save the result to a FITS table
if os.path.exists('splinefit-1d.fits'):
	os.unlink('splinefit-1d.fits')
# this file can later be loaded and evaluated with the photospline C library
splinefitstable.write(result, 'splinefit-1d.fits')

# Plot the result
import pylab
from utils import colorize
pylab.scatter(x1, z, label='Data')

# Plot individual basis splines 
xfine = numpy.linspace(knots[0][0], knots[0][-1], 10001)
splines = [numpy.array([bspline(knots[0], x, n, order) for x in xfine]) for n in range(0,len(knots[0])-2-1)]
for c, n in colorize(range(len(splines))):
	pylab.plot(xfine, result.coefficients[n]*splines[n], color=c)
# Plot the spline surface (sum of all the basis functions)
pylab.plot(xfine, glam.grideval(result, [xfine]), label='Spline fit', color='k')

pylab.legend(loc='upper left')
pylab.show()
