
import numpy
import copy

def colorize(sequence, cmap='jet'):
	from matplotlib.colors import Normalize
	from matplotlib import cm
	norm = Normalize(vmin=0, vmax=len(sequence)-1)
	cmap = getattr(cm, cmap)
	for i, s in enumerate(sequence):
		yield cmap(norm(i)), s

def pad_knots(knots, order=2):
	"""
	Pad B-spline knots out for full support at the boundaries
	"""
	pre = knots[0] - (knots[1]-knots[0])*numpy.arange(order, 0, -1)
	post = knots[-1] + (knots[-1]-knots[-2])*numpy.arange(1, order+1)
	return numpy.concatenate((pre, knots, post))

def stack_tables(tablist, order = 2):
	"""
	Given N-dimensional splines surfaces evenly spaced in a new
	dimension (N+1), construct an N+1-dimensional spline that
	maximally fills the convex hull given by the input surfaces.
	
	:param tablist: list of (splinetable, coordinate) tuples
	:param order: order of the interpolating spline
	"""
	
	bigtab = None

	for slice, position in tablist:
		
		coefficients = slice.coefficients.reshape( \
		    slice.coefficients.shape + (1,))
		ndim = coefficients.ndim
		
		if bigtab is None:
			bigtab = copy.deepcopy(slice)
			bigtab.coefficients = coefficients
			bigtab.knots.append([position])
			try:
				bigtab.order.append(order)
			except AttributeError:
				bigtab.order = [bigtab.order, order]
			bigtab.periods = numpy.append(bigtab.periods, 0.)
		else:
			bigtab.knots[ndim - 1].append(position)
			bigtab.coefficients = numpy.concatenate(
			    (bigtab.coefficients, coefficients),
			    ndim - 1)
	
	# Shift the knots so that the maxima of the splines along
	# the new dimension are at the positions of the sub-surfaces
	baseknots = numpy.asarray(bigtab.knots[ndim - 1])
	baseknots += (baseknots.max()-baseknots.min())/(2.0*baseknots.size)*(order-1)
	bigtab.knots[ndim - 1] = pad_knots(baseknots, order)[:-1]
	
	return bigtab
